#ifndef CASEBASE_H
#define CASEBASE_H

#include "CaseRetriever.h"
#include "Case.h"

#include <map>
#include <set>
#include <string>
#include <vector>
#include <typeinfo>
#include <cassert>

namespace cppcbr {

class AttributeSetCase;
class AttributeValue;

class ValueSimilarityBase {
public:
	virtual ~ValueSimilarityBase() {}
	virtual double operator()(const AttributeValue* from, const AttributeValue* to) = 0;
};

template <typename T> class ValueSimilarity: public ValueSimilarityBase {
public:
	ValueSimilarity(double (*function)(const T& a, const T& b)): function(function) {}
	double operator()(const AttributeValue* from, const AttributeValue* to) {
		const ClassAttributeValue<T>* fromv = dynamic_cast<const ClassAttributeValue<T>*>(from);
		const ClassAttributeValue<T>* tov = dynamic_cast<const ClassAttributeValue<T>*>(to);
		if (!fromv || !tov) {
			throw string("Similarity error: Values are not of type ") + typeid(T).name() + ".";
		}
		return function(fromv->getValue(), tov->getValue());
	}
private:
	double (*function)(const T& a, const T& b);
};

class SimilarityEvaluator {
public:
	virtual ~SimilarityEvaluator() {
		for (map<string, ValueSimilarityBase*>::iterator it = similarityEvaluators.begin(); it != similarityEvaluators.end(); ++it) {
			delete it->second;
		}
	}
	virtual double evaluateSimilarity(string name, const AttributeValue* from, const AttributeValue* to) const {
		map<string, ValueSimilarityBase*>::const_iterator it = similarityEvaluators.find(name);
		if (it == similarityEvaluators.end()) {
			throw string("No similarity evaluator defined for feature \"") + name + "\".";
		}
		assert(name == it->first);
		ValueSimilarityBase* vsb = it->second;
		return vsb->operator()(from, to);
	}
	virtual void addSimilarityMeasure(string name, ValueSimilarityBase* vsb) {
		if (vsb == NULL) {
			similarityEvaluators.erase(name);
		} else {
			similarityEvaluators[name] = vsb;
		}
	}
private:
	map<string, ValueSimilarityBase*> similarityEvaluators;
};

template <typename C, typename CR> class RetrievalCaseBase: public SimilarityEvaluator {
public:
	typedef C case_type;
	typedef set<case_type*> casebase_type;
	typedef vector<case_type*> retrieved_type;
	RetrievalCaseBase(): caseRetriever() {}
	virtual ~RetrievalCaseBase() {
		for (typename casebase_type::iterator it = caseBase.begin(); it != caseBase.end(); ++it) {
			delete *it;
		}
	}
	virtual vector<C*> retrieveSimilarCases(const C* c) const {
		return caseRetriever(c, caseBase, this);
	}
	virtual void insertCase(C* c) {
		caseBase.insert(c);
	}
private:
	casebase_type caseBase;
	CR caseRetriever;
};

class SimpleCaseBase: public RetrievalCaseBase<AttributeSetCase, CaseRetriever<AttributeSetCase> > {
};

}

#endif
