#include "common.h"
#include "GoBoard.h"
#include "tactics.h"

#include <algorithm>

namespace go {

void GoBoard::build_groups() {
	some_group_in_atari = false;
	groups.clear();
	groups.push_back(group(0));
	memset(group_idx, 0, sizeof(group_idx));
	bool visited[board_size][board_size];
	memset(visited, 0, sizeof(visited));
	for (int y = board_size - 1; y >= 0; --y) {
		for (int x = 0; x < board_size; ++x) {
			if (visited[y][x]) continue;
			position p(x, y);
			int player = getTile(p);
			if (!player) continue;
			group g(player);
			int offx[] = { -1, +1, 0, 0 };
			int offy[] = { 0, 0, -1, +1 };
			deque<position> q;
			q.push_back(position(x, y));
			while (q.size()) {
				position p = q.front();
				q.pop_front();
				g.add_stone(p);
				group_idx[p.y][p.x] = groups.size();
				visited[p.y][p.x] = true;
				for (int i = 0; i < 4; ++i) {
					int nx = p.x + offx[i];
					int ny = p.y + offy[i];
					if (nx < 0 || nx >= board_size) continue;
					if (ny < 0 || ny >= board_size) continue;
					if (visited[ny][nx]) continue;
					position np(nx, ny);
					int nt = getTile(np);
					if (nt == g.player) {
						q.push_back(np);
						visited[ny][nx] = true;
					}
					else if (nt == 0) g.add_liberty(np);
				}
			}
			groups.push_back(g);
			if (g.num_liberties <= 1) some_group_in_atari = true;
		}
	}
}

struct position_neighbours {
	position tile[4];
	int count;
	position_neighbours(const position& pos): count(0) {
		if (pos.x > 0)
			tile[count++] = position(pos.x - 1, pos.y);

		if (pos.x < (board_size - 1))
			tile[count++] = position(pos.x + 1, pos.y);

		if (pos.y > 0)
			tile[count++] = position(pos.x, pos.y - 1);

		if (pos.y < (board_size - 1))
			tile[count++] = position(pos.x, pos.y + 1);
	}
};

struct group_neighbours {
	int group[4];
	int count;
	group_neighbours(const GoBoard& board, const position& pos): count(0) {
		position_neighbours pn(pos);
		for (int i = 0; i < pn.count; ++i) {
			if (board.getTile(pn.tile[i])) {
				int g = board.group_idx[pn.tile[i].y][pn.tile[i].x];
				int j = 0;
				for (j = 0; j < count; ++j) {
					if (group[j] == g) break;
				}
				if (j == count) {
					group[j] = g;
					count++;
				}
			}
		}
	}
};

int GoBoard::isCapturingMove(const position& pos, int player) const {
	if (!some_group_in_atari) return 0;
	int opponent = 3 - player;
	group_neighbours gn(*this, pos);
	int captures = 0;
	for (int i = 0; i < gn.count; ++i) {
		const group& g = groups[gn.group[i]];
		if (g.player == opponent && g.num_liberties <= 1) {
			captures += g.num_stones;
		}
	}
	return captures;
}

bool GoBoard::isSelfAtari(const position& p, int player) const {
	int liberties = 0;
	position_neighbours pn(p);
	for (int i = 0; i < pn.count; ++i) {
		if (!getTile(pn.tile[i])) liberties++;
	}
	if (liberties > 1) return false;
	group_neighbours gn(*this, p);
	for (int i = 0; i < gn.count; ++i) {
		const group& g = groups[gn.group[i]];
		if (g.player == player) {
			liberties += g.num_liberties - 1;
		}
	}
	return (liberties <= 1);
}

void GoBoard::play(const position& pos) {
	if (pos.is_pass()) {
		pass();
		return;
	}
	assert(!game_over);
	++num_plays;
	++simuls;
	int opponent = 3 - player;
	group_neighbours gn(*this, pos);
	bool must_rebuild = false;
	for (int i = 0; i < gn.count; ++i) {
		group& g = groups[gn.group[i]];
		if (g.player == opponent && g.num_liberties <= 1) {
			for (int j = 0; j < g.num_stones; ++j) {
				const position& kp = g.stones_pos[j];
				setTile(kp, 0);
				group_idx[kp.y][kp.x] = 0;
				if (player == 1) black_captures++;
				else if (player == 2) white_captures++;
			}
			must_rebuild = true;
			g.num_stones = 0;
		}
	}
	setTile(pos, player);
	position_neighbours pn(pos);
	for (int i = 0; i < pn.count; ++i) {
		if (getTile(pn.tile[i])) must_rebuild = true;
	}
	if (must_rebuild) {
		build_groups();
	} else {
		group g(player);
		g.add_stone(pos);
		for (int i = 0; i < pn.count; ++i) {
			g.add_liberty(pn.tile[i]);
		}
		groups.push_back(g);
	}

	player = opponent;
	last_play = pos;
	visited_boards.insert(hash_code);
}

bool GoBoard::isLegalHashPlay(const position& p) const {
	if (isCapturingMove(p, player)) {
		int opponent = 3 - player;
		unsigned int new_hash = hash_code;
		new_hash ^= hash_values[p.y][p.x][0];
		new_hash ^= hash_values[p.y][p.x][player];
		group_neighbours gn(*this, p);
		for (int i = 0; i < gn.count; ++i) {
			const group& g = groups[gn.group[i]];
			if (g.player == opponent && g.num_liberties <= 1) {
				for (int j = 0; j < g.num_stones; ++j) {
					const position& sp = g.stones_pos[j];
					new_hash ^= hash_values[sp.y][sp.x][opponent];
					new_hash ^= hash_values[sp.y][sp.x][0];
				}
			}
		}
		return !visited_boards.count(new_hash);
	}
	unsigned int new_hash = hash_code;
	new_hash ^= hash_values[p.y][p.x][getTile(p)];
	new_hash ^= hash_values[p.y][p.x][player];
	return !visited_boards.count(new_hash);
}

bool GoBoard::isLegalMove(const position& p, int player) const {
	if (getTile(p)) return false;
	// Empty surrounding tiles?
	position_neighbours pn(p);
	for (int i = 0; i < pn.count; ++i) {
		if (!getTile(pn.tile[i])) return true;
	}
	// Connected to live group?
	group_neighbours gn(*this, p);
	for (int i = 0; i < gn.count; ++i) {
		const group& g = groups[gn.group[i]];
		if (g.player == player && g.num_liberties > 1) return true;
	}
	if (isCapturingMove(p, player)) return true;
	return false;
}

bool GoBoard::isConnectSplitMove(const position& p, int player) const {
	// Connecting or splitting groups
	group_neighbours gn(*this, p);
	int opponent = 3 - player;
	int mine = 0;
	int opps = 0;
	for (int i = 0; i < gn.count; ++i) {
		const group& g = groups[gn.group[i]];
		if (g.player == player) mine++;
		if (g.player == opponent) opps++;
	}
	if (mine >= 2 || opps >= 2) return true;
	return false;
}

bool GoBoard::isInterestingMove(const position& p, int player) const {
	if (!isLegalMove(p, player)) return false;
	// Empty surrounding tiles?
	position_neighbours pn(p);
	for (int i = 0; i < pn.count; ++i) {
		if (!getTile(pn.tile[i])) return true;
	}
	if (isConnectSplitMove(p, player)) return true;
	if (isCapturingMove(p, player)) return true;
	return false;
}

bool GoBoard::isFinalEndgameMove(const position& p, int player) const {
	if (!isLegalMove(p, player)) return false;
	// Is dame point?
	group_neighbours gn(*this, p);
	if (gn.count > 1) return true;
	return false;
}

position get_capture_move(const GoBoard& board, bool only_big) {
	int min_stones = 1;
	if (only_big) min_stones = 2;
	vector<position> both;
	vector<position> capture;
	vector<position> defend;
	for (int x = 0; x < board_size; ++x) {
		for (int y = 0; y < board_size; ++y) {
			position p(x, y);
			if (board.getTile(p)) continue;
			bool cap = (board.isCapturingMove(p, board.player) >= min_stones);
			bool def = (board.isCapturingMove(p, 3 - board.player) >= min_stones);
			if (cap && def && board.isLegalPlay(p)) {
				both.push_back(p);
			} else if (def && !board.isSelfAtari(p, board.player) && board.isLegalPlay(p)) {
				defend.push_back(p);
			} else if (cap && board.isLegalPlay(p)) {
				capture.push_back(p);
			}
		}
	}
	if (both.size()) return random_element(both);
	if (defend.size()) return random_element(defend);
	if (capture.size()) return random_element(capture);
	return position(-1, -1);
}

position GoBoard::get_funky_move() const {
	vector<position> connsplits;
	vector<position> randoms;
	for (int x = 0; x < board_size; ++x) {
		for (int y = 0; y < board_size; ++y) {
			position p(x, y);
			if (getTile(p)) continue;
			if (isInterestingMove(p, player)) {
				if (!isSelfAtari(p, player)) {
					if (isLegalPlay(p)) {
						randoms.push_back(p);
						if (isConnectSplitMove(p, player)) {
							connsplits.push_back(p);
						}
					}
				}
			}
		}
	}
	vector<position> funs;
	for (unsigned int i = 0; i < randoms.size(); ++i) {
		position p = randoms[i];
		if (p.x == 2 || p.x == 3 || p.x == board_size - 3 || p.x == board_size - 4) {
			if (p.y == 2 || p.y == 3 || p.y == board_size - 3 || p.y == board_size - 4) {
				bool fun = true;
				for (int y = p.y - 1; y <= p.y + 1; ++y) {
					for (int x = p.x - 1; x <= p.x + 1; ++x) {
						position pn(x, y);
						if (getTile(pn) && pn != p) fun = false;
					}
				}
				if (fun) {
					funs.push_back(p);
				}
			}
		}
	}
	if (funs.size()) return random_element(funs);
	if (connsplits.size()) return random_element(connsplits);
	if (randoms.size()) return random_element(randoms);
	vector<position> endgames;
	for (int x = 0; x < board_size; ++x) {
		for (int y = 0; y < board_size; ++y) {
			position p(x, y);
			if (getTile(p)) continue;
			if (isFinalEndgameMove(p, player)) {
				if (isLegalPlay(p)) {
					endgames.push_back(p);
				}
			}
		}
	}
	if (endgames.size()) return random_element(endgames);
	return position(-1, -1);
}

double play_funky(const GoBoard& input_board) {
	GoBoard board(input_board);
	position old_pos = board.get_last_play();
	while (!board.is_game_over()) {
		position p = get_capture_move(board, true);
		if (p.is_pass()) p = get_capture_move(board, false);
		if (p.is_pass()) p = board.get_funky_move();
		board.play(p);
		old_pos = p;
	}
	double gamescore = board.get_score();
	return gamescore;
}

double play_some_funky_and_eval(GoBoard& board) {
	position old_pos = board.get_last_play();
// 	cout << "started at " << board.num_plays << flush;
	while (!board.is_game_over()) {
		position p = get_capture_move(board, true);
		if (p.is_pass()) p = get_capture_move(board, false);
		if (p.is_pass()) p = board.get_funky_move();
		if (!board.isLegalPlay(p)) {
			cout << "Error!" << endl;
			board.print();
			cout << "and tried to play " << p << " !!!" << endl;
			cout << "get_cap_big: " << get_capture_move(board, true) << endl;
			cout << "get_cap_small: " << get_capture_move(board, false) << endl;
			cout << "get_funk: " << board.get_funky_move() << endl;
		}
		assert(board.isLegalPlay(p));
		board.play(p);
		old_pos = p;
	}
// 	cout << " ended at " << board.num_plays << endl;
	if (board.is_game_over()) {
		return board.get_score();
	} else {
		return board.estimate_score();
	}
}

void fill_empties(const GoBoard& board, vector<position>& empties) {
	empties.clear();
	for (int y = 0; y < board_size; ++y) {
		for (int x = 0; x < board_size; ++x) {
			position p(x, y);
			if (board.getTile(p) == 0) {
				empties.push_back(p);
			}
		}
	}
}

struct empty_area {
	vector<position> tiles;
	bool black_nearby;
	bool white_nearby;
	empty_area(): black_nearby(false), white_nearby(false) {}
};

double GoBoard::eval_randomed_board() {
	vector<empty_area> eas;
	eas.push_back(empty_area());

	bool visited[board_size][board_size];
	memset(visited, 0, sizeof(visited));
	for (int y = board_size - 1; y >= 0; --y) {
		for (int x = 0; x < board_size; ++x) {
			if (visited[y][x]) continue;
			position p(x, y);
			if (getTile(p) == 0) {
				empty_area ea;
				deque<position> q;
				q.push_back(p);
				visited[y][x] = true;
				while (q.size()) {
					position p = q.front();
					q.pop_front();
					ea.tiles.push_back(p);
					position_neighbours pns(p);
					for (int i = 0; i < pns.count; ++i) {
						const position& pn = pns.tile[i];
						int t = getTile(pn);
						if (t == 0 && !visited[pn.y][pn.x]) {
							q.push_back(pn);
							visited[pn.y][pn.x] = true;
						} else if (t == 1) {
							ea.black_nearby = true;
						} else if (t == 2) {
							ea.white_nearby = true;
						}
					}
				}
				eas.push_back(ea);
			}
		}
	}

	int eas_map[board_size][board_size];
	memset(eas_map, 0, sizeof(eas_map));

	for (unsigned int i = 1; i < eas.size(); ++i) {
		for (unsigned int j = 0; j < eas[i].tiles.size(); ++j) {
			const position& p = eas[i].tiles[j];
			eas_map[p.y][p.x] = i;
		}
	}

	set<int> map_eyespaces_group[eas.size()];

	bool alive[groups.size()];
	bool dead[groups.size()];
	memset(alive, 0, sizeof(alive));
	memset(dead, 0, sizeof(dead));

	int first_group_eyespace[groups.size()];

	for (unsigned int i = 1; i < groups.size(); ++i) {
		int first_area = -1;
		bool got_many = false;
		const group& g = groups[i];
		for (int j = 0; j < g.num_liberties; ++j) {
			const position& lp = g.liberties_pos[j];
			int eas_num = eas_map[lp.y][lp.x];
			assert(eas_num);
			if (eas_num != first_area) {
				if (first_area == -1) first_area = eas_num;
				else got_many = true;
			}
			map_eyespaces_group[eas_num].insert(i);
		}
		first_group_eyespace[i] = first_area;
		if (got_many) alive[i] = true;
	}

	for (unsigned int i = 1; i < groups.size(); ++i) {
		if (alive[i]) continue;
		int eyespace_idx = first_group_eyespace[i];
		const set<int>& groups_there = map_eyespaces_group[eyespace_idx];
		int eyespace_size = eas[eyespace_idx].tiles.size();
		if (groups_there.size() == 1) {
			if (eyespace_size < 6) {
				dead[i] = true;
			}
		}
	}

	for (unsigned int i = 1; i < groups.size(); ++i) {
		if (alive[i]) continue;
		if (dead[i]) continue;
		const group& g = groups[i];
		int eyespace_idx = first_group_eyespace[i];
		const set<int>& groups_there = map_eyespaces_group[eyespace_idx];
		bool saved = false;
		for (set<int>::const_iterator it = groups_there.begin(); it != groups_there.end(); ++it) {
			const group& itg = groups[*it];
			if (itg.player == g.player) {
				if (alive[*it]) saved = true;
			}
		}
		if (saved) alive[i] = true;
		else dead[i] = true;
	}

	for (unsigned int i = 1; i < groups.size(); ++i) {
		if (dead[i]) {
			const group& g = groups[i];
			for (int j = 0; j < g.num_stones; ++j) setTile(g.stones_pos[j], 0);
			if (g.player == 1) white_captures += g.num_stones;
			if (g.player == 2) black_captures += g.num_stones;
		}
	}

	return get_board_tile_score() + black_captures - white_captures - komi;
}

void GoBoard::turbo_rebuild_groups() {
	groups.clear();
	groups.reserve(20);
	groups.push_back(group(0));
	bool visited[board_size][board_size];
	memset(visited, 0, sizeof(visited));
	position dq[board_size * board_size];
	int dq_pos = 0;
	int dq_idx = 0;
	for (int y = board_size - 1; y >= 0; --y) {
		for (int x = 0; x < board_size; ++x) {
			if (visited[y][x]) continue;
			position p(x, y);
			int player = getTile(p);
			if (!player) continue;
			group g(player);
			dq[dq_pos++] = position(x, y);
			int offx[] = { -1, +1, 0, 0 };
			int offy[] = { 0, 0, -1, +1 };
			while (dq_idx < dq_pos) {
				const position& p = dq[dq_idx++];
				g.add_stone(p);
				visited[p.y][p.x] = true;
				for (int i = 0; i < 4; ++i) {
					int nx = p.x + offx[i];
					int ny = p.y + offy[i];
					if (nx < 0 || nx >= board_size) continue;
					if (ny < 0 || ny >= board_size) continue;
					if (visited[ny][nx]) continue;
					position np(nx, ny);
					int nt = getTile(np);
					if (nt == g.player) {
						dq[dq_pos++] = np;
						visited[ny][nx] = true;
					}
					else if (nt == 0) g.num_liberties++;
				}
			}
			groups.push_back(g);
		}
	}
}

double GoBoard::turbo_play() {
	vector<position> empties;
	empties.reserve(board_size * board_size);
	fill_empties(*this, empties);
	int max_moves = empties.size();
// 	if (max_moves <= board_size * 2) return play_some_funky_and_eval(*this);
	random_shuffle(empties.begin(), empties.end());
	for (int i = 0; i < max_moves; ++i) {
		setTile(empties[i], player);
		player = 3 - player;
	}
	num_plays += max_moves;

	turbo_rebuild_groups();

	// Removing small dead groups, without counting captures
	for (unsigned int i = 0; i < groups.size(); ++i) {
		const group& g = groups[i];
		if (g.num_stones <= 3) {
			for (int j = 0; j < g.num_stones; ++j) {
				setTile(g.stones_pos[j], 0);
			}
		}
	}

	turbo_rebuild_groups();

	// Removing remaining dead groups, without counting captures
	for (unsigned int i = 0; i < groups.size(); ++i) {
		const group& g = groups[i];
		if (g.num_liberties == 0) {
			for (int j = 0; j < g.num_stones; ++j) {
				setTile(g.stones_pos[j], 0);
			}
		}
	}

	build_groups();

	last_play = position(-1, -1);

	return eval_randomed_board();
// 	return play_some_funky_and_eval(*this);
}

double play_purely_random(GoBoard& board) {
	while (!board.is_game_over()) {
		vector<position> endgames;
		for (int x = 0; x < board_size; ++x) {
			for (int y = 0; y < board_size; ++y) {
				position p(x, y);
				if (board.getTile(p)) continue;
				if (board.isInterestingMove(p, board.player) || board.isFinalEndgameMove(p, board.player)) {
					if (board.isLegalPlay(p)) {
						endgames.push_back(p);
					}
				}
			}
		}
		if (endgames.size()) {
			position p = random_element(endgames);
			board.play(p);
		} else {
			board.pass();
		}
	}
	return board.get_score();
}

double do_play(const GoBoard& input_board) {
	GoBoard board(input_board);
// 	cout << "do_play()" << flush;
// 	return play_purely_random(board);
	if (board_size == 19 && input_board.num_plays < 40) {
		return board.turbo_play();
	}
	return play_some_funky_and_eval(board);
// 	double score = play_some_funky_and_eval(input_board);
// 	cout << " = " << score << endl;
// 	return score;
}

vector<position> GoBoard::getInterestingPlays() const {
	vector<position> posses;
	for (int x = 0; x < board_size; ++x) {
		for (int y = 0; y < board_size; ++y) {
			position p(x, y);
			if (isInterestingMove(p, player) || isFinalEndgameMove(p, player)) {
				if (isLegalHashPlay(p)) {
					posses.push_back(p);
				}
			}
		}
	}
	return posses;
}

bool GoBoard::isLegalPlay(const position& p) const {
	if (p.is_pass()) return true;
	return (isLegalMove(p, player) && isLegalHashPlay(p));
}

double GoBoard::get_board_estimate_tile_score() const {
	return get_board_score_estimate(*this);
}

double GoBoard::get_board_tile_score() const {
	double score = 0;
	for (int y = 0; y < board_size; ++y) {
		for (int x = 0; x < board_size; ++x) {
			int t = getTile(position(x, y));
			if (t == 1) {
				score++;
			} else if (t == 2) {
				score--;
			}
		}
	}
	bool visited[board_size][board_size];
	memset(visited, 0, sizeof(visited));
	for (int y = 0; y < board_size; ++y) {
		for (int x = 0; x < board_size; ++x) {
			position initp(x, y);
			if (getTile(initp) == 0 && !visited[y][x]) {
				bool reached_black = false;
				bool reached_white = false;
				visited[y][x] = true;
				deque<position> q;
				q.push_back(initp);
				while (q.size()) {
					position p = q.front();
					q.pop_front();
					position_neighbours pn(p);
					for (int i = 0; i < pn.count; ++i) {
						int t = getTile(pn.tile[i]);
						if (t == 0 && !visited[pn.tile[i].y][pn.tile[i].x]) {
							q.push_back(pn.tile[i]);
							visited[pn.tile[i].y][pn.tile[i].x] = true;
						} else if (t == 1) {
							reached_black = true;
						} else if (t == 2) {
							reached_white = true;
						}
					}
				}
				if (reached_black && reached_white) {
				} else if (reached_black) {
					score++;
				} else if (reached_white) {
					score--;
				}
			}
		}
	}
	return score;
}

void GoBoard::print() const {
	cout << "lastplay: " << get_last_play() << endl;
	cout << " {board} (move " << (num_plays+1) << ")" << endl;
	cout << ((player==1)?"X":"O") << " to play (" << hex << hash_code << dec << ")" << endl;
	cout << "  ";
	for (int x = 0; x < board_size; ++x) {
		char c = char('A' + x);
		if (c >= 'I') c++;
		cout << " " << c << " ";
	}
	cout << endl;
	for (int y = board_size - 1; y >= 0; --y) {
		if (y + 1 < 10) cout << " ";
		cout << (y + 1);
		for (int x = 0; x < board_size; ++x) {
			string s = " . ";
// 			if (!b.isLegalPlay(position(x, y))) s = " _ ";
			if (board_size == 19) {
				if (x == 3 || x == 9 || x == 15) {
					if (y == 3 || y == 9 || y == 15) {
						s = " + ";
					}
				}
			} else if (board_size == 9) {
				if (x == 2 || x == 4 || x == 6) {
					if (y == 2 || y == 4 || y == 6) {
						s = " + ";
					}
				}
			}
			if (getTile(position(x, y)) == 1) {
				s = " X ";
				if (get_last_play() == position(x, y)) s = "(X)";
			} else if (getTile(position(x, y)) == 2) {
				s = " O ";
				if (get_last_play() == position(x, y)) s = "(O)";
			}
			cout << s;
		}
		cout << " " << (y + 1) << endl;
	}
	cout << "  ";
	for (int x = 0; x < board_size; ++x) {
		char c = char('A' + x);
		if (c >= 'I') c++;
		cout << " " << c << " ";
	}
	cout << endl;
	cout << endl;


// 	for (int y = board_size - 1; y >= 0; --y) {
// 		for (int x = 0; x < board_size; ++x) {
// 			if (group_idx[y][x])
// 				cout << char('A' - 1 + group_idx[y][x]) << " ";
// 			else
// 				cout << ". ";
// 		}
// 		cout << endl;
// 	}
// 	cout << endl << endl;
}

}
