#include "common.h"
#include "UCT.h"
#include "extra.h"

#include <bitset>

namespace go {

/*

one episode = one rollout play
action a = move at position p

to give score for a at s, we average the returns of all episodes where a was played, not only immediately played

Qrave(s, a) = RAVEst for a in s
after each episodes s1,a1,s2,a2,...,sT, the action values are updated for
  every state st1 in T and every subsequent action at2 such that
    at2 in A(st1), t1 < t2 and for all t < t2, at != at2,

m(st1, at2) <- m(st1, at2) + 1
qrave(st1, at2) <- qrave(st1, at2) + 1 / m(st1, at2)[Rt1 - qrave(st1, at2)]

m(s, a) counts the number of times a has been selected following state s

q+rave(s, a) = qrave(s, a) + c sqrt(log m(s) / m(s, a))
beta(s, a) = sqrt(k / (3n(s) + k))
q+ur(s, a) = beta(s, a) q+rave(s, a) + (1 - beta(s, a)) q+uct(s, a)
piUR(s) = argmax{a} q+ur(s, a)

where m(s) = sum{a} m(s, a)

k = equiv parameter = 1000

*/

double get_uct_score(const UCT_node* n, const UCT_node* parent) {
	double uctscore;
	if (!n) {
		assert(parent);
		double winrate = double(parent->plays - parent->wins) / parent->plays;
		double uct = sqrt(log(parent->plays) / 5.0);
		uctscore = winrate + uct;
	} else if (!n->parent) {
		throw string("How did I get here? v1");
	} else if (n->plays == 0) {
		throw string("How did I get here? v2");
	} else {
		double winrate = double(n->wins) / n->plays;
		double uct = sqrt(log(n->parent->plays) / (5 * n->plays));
		uctscore = winrate + uct;
	}
	return uctscore;
}

double get_decision_score(UCT_node* n) {
	if (!n) return -1;
	return double(n->wins) / n->plays;
}

int UCT::get_best_child_idx(UCT_node* n, const GoBoard& board) const {
	double bestScore = -1000;
	int bestidx = -1;
	for (unsigned int i = 0; i < n->children.size(); ++i) {
		const pair<position, UCT_node*> it = n->children[i];
		double uctscore = get_uct_score(it.second, n);
		if (using_cbr && board.num_plays <= 20) uctscore += get_cbr_bias(it.first, board);
		if (uctscore > bestScore) {
			bestScore = uctscore;
			bestidx = i;
		}
	}
// 	cout << "Picking " << n->children[bestidx].first << endl;
	return bestidx;
}

position UCT::get_best_move() const {
	position cbr_pos = get_cbr_closest_position(root_board);
	if (using_cbr) {
		cout << "CBR bias position for root node[" << std::hex << root_board.hash_code << std::dec << "]: " << cbr_pos << endl;
	}
	const UCT_node* node = getRootNode();
	double best_score = -1000;
	double cbr_score = -1000;
	vector<pair<position, UCT_node*> > best;
	for (unsigned int i = 0; i < node->children.size(); ++i) {
		pair<position, UCT_node*> it = node->children[i];
		if (it.second) {
			double score = get_decision_score(it.second);
			if (using_cbr && root_board.num_plays <= 20) score += get_cbr_bias(it.first, root_board);
			if (score > best_score && it.second->plays >= 10) {
				best_score = score;
				best.clear();
				best.push_back(it);
			} else if (score == best_score) {
				best.push_back(it);
			}
			if (it.first == cbr_pos) {
				cbr_score = score;
			}
		}
	}
	if (!best.size()) { // No nodes were found that had at least 10 plays
		for (unsigned int i = 0; i < node->children.size(); ++i) {
			pair<position, UCT_node*> it = node->children[i];
			if (it.second) {
				double score = get_decision_score(it.second);
				if (using_cbr && root_board.num_plays <= 20) score += get_cbr_bias(it.first, root_board);
				if (score > best_score) {
					best_score = score;
					best.clear();
					best.push_back(it);
				} else if (score == best_score) {
					best.push_back(it);
				}
			}
		}
	}
	if (using_cbr && cbr_score >= 0) {
		cout << "CBR-suggested move " << cbr_pos << " is worth " << cbr_score << " (based on " << get_cbr_source_filename(root_board) << ")" << endl;
	}
	cout << "Computer considers best move " << best.front().first << " worth " << best_score << endl;
	if (best.front().first.is_pass()) {
// 		cout << "  hmpf passing!!" << endl;
		GoBoard board(root_board);
		deque<position> moves;
		UCT_node* n = node->parent;
		while (n) {
			UCT_node* p = n->parent;
			for (unsigned int i = 0; i < p->children.size(); ++i) {
				if (p->children[i].second == n) {
					moves.push_front(p->children[i].first);
				}
			}
			n = n->parent;
		}
		for (unsigned int i = 0; i < moves.size(); ++i) {
			board.play(moves[i]);
		}
// 		board.print();
// 		cout << " intmoves: " << board.getInterestingPlays().size() << endl;
	}
	if (best_score < 0.01) {
// 		return make_pair(position(-2, -2), (UCT_node*)NULL);
	}
	return best.front().first;
}

void print_sub_node_tree(const UCT_node* node, const GoBoard& board, string prefix = "", int recurse_limit = 3) {
	vector<pair<double, pair<position, UCT_node*> > > children;
	for (unsigned int i = 0; i < node->children.size(); ++i) {
		pair<position, UCT_node*> it = node->children[i];
		double score = get_decision_score(it.second);
		if (using_cbr && board.num_plays <= 20) score += get_cbr_bias(it.first, board);
		children.push_back(make_pair(-score, it));
	}
	sort(children.begin(), children.end());
	for (int i = 0; i < (recurse_limit * 2 + 1) && i < (int)children.size(); ++i) {
		double score = -children[i].first;
		pair<position, UCT_node*> it = children[i].second;
		cout << prefix << it.first << ": ";
		if (it.second) {
			cout << score << " (" << it.second->wins << " / " << it.second->plays << ")";
		} else {
			cout << "??? (unplayed)";
		}
		cout << " uct[" << get_uct_score(it.second, node) << "]";
		if (using_cbr && board.num_plays <= 20) cout << " cbr[" << get_cbr_closest_position(board) << ": " << get_cbr_bias(it.first, board) << "  " << get_cbr_source_filename(board) << "]";
		cout << endl;
		if (it.second && i < recurse_limit) {
			GoBoard b(board);
			b.play(it.first);
			print_sub_node_tree(it.second, b, prefix + " ", recurse_limit - 1 - i/2);
		}
	}
}

void UCT::print_node_tree() const {
	print_sub_node_tree(getRootNode(), root_board, "", 3);
}

// 		if atari then
// 			if possible to save
// 				= random save
// 		if there is an interesting move nearby
// 			= random interesting move nearby
// 		if there is a capturing move
// 			= random capturing move
// 		= random move
// 		hane patterns: { "BWB . xxx", "BW  . x x", "BWxB. x x", "BWW b x x" }
// 		cut pattern: "BWxW xxxx" and not ("BWxW.Wx x" or "BWxW. xWx")
// 		             "xBxW.Wmmm"
// 		side patterns: "B xW.x" <-- suspicious!
// 		               "xBxb.W", "xBWxbx", "xWBxbw", "xWBBbw"
bool single_pattern_match(const char region[9], const char* pattern) {
	for (int i = 0; i < 9; ++i) {
		if (pattern[i] == 'x') continue;
		if (pattern[i] == '.') continue;
		if (pattern[i] != region[i]) return false;
	}
	return true;
}

bool pattern_match(const GoBoard& board, const position p) {
	const char* hane_patterns[] = { "BWB . xxx", "BW  . x x", "BWxB. x x", "BWW b x x" };
	const char* cut_pattern1[] = { "BWxW xxxx", "BWxW.Wx x", "BWxW. xWx" };
//	char* cut_pattern2[] = { "xBxW.Wmmm" };
//	char* side_patterns[] = { "B xW.x", "xBxb.W", "xBWxbx", "xWBxbw", "xWBBbw" };
	bool on_side = false;
	if (p.x == 0 || p.y == 0 || p.x == board_size - 1 || p.y == board_size - 1) {
		on_side = true;
		return false;
	}
	if (!on_side) {
		char region[9];
		for (int y = 0; y <= 2; ++y) {
			for (int x = 0; x <= 2; ++x) {
				position rp(p.x - 1 + x, p.y - 1 + y);
				char c = ' ';
				if (board.getTile(rp) == 1) c = 'B';
				else if (board.getTile(rp) == 2) c = 'W';
				region[y * 3 + x] = c;
			}
		}
		for (int swap = 0; swap < 2; ++swap) {
			for (int rot = 0; rot < 4; ++rot) {
				for (unsigned int i = 0; i < 4; ++i) {
					const char* pattern = hane_patterns[i];
					if (single_pattern_match(region, pattern)) return true;
				}
				if (single_pattern_match(region, cut_pattern1[0])) {
					if (!single_pattern_match(region, cut_pattern1[1]) && !single_pattern_match(region, cut_pattern1[2])) return true;
				}
				char tmp = region[0];
				region[0] = region[2];
				region[2] = region[8];
				region[8] = region[6];
				region[6] = tmp;
				tmp = region[1];
				region[1] = region[5];
				region[5] = region[7];
				region[7] = region[3];
				region[3] = tmp;
			}
			for (int i = 0; i < 9; ++i) {
				if (region[i] == 'B') region[i] = 'W';
				else if (region[i] == 'W') region[i] = 'B';
			}
		}
	}
	return false;
}

position get_pattern_move(GoBoard& board, const position& old_pos) {
	vector<position> plays;
	if (!old_pos.is_pass()) {
		for (int x = old_pos.x - 1; x < old_pos.x + 1; ++x) {
			for (int y = old_pos.y - 1; y < old_pos.y + 1; ++y) {
				position p(x, y);
				if (board.getTile(p)) continue;
				if (board.isLegalPlay(p)) {
					if (pattern_match(board, p)) {
						plays.push_back(p);
					}
				}
			}
		}
	}
	if (!plays.size()) {
		return position(-1, -1);
	}
	return random_element(plays);
}

}
