#ifndef UCT_H
#define UCT_H

#include "common.h"
#include "GoBoard.h"

namespace go {

struct UCT_node {
	UCT_node* parent;
	double wins;
	int plays;
	vector<pair<position, UCT_node*> > children;
	UCT_node(UCT_node* parent): parent(parent), wins(0), plays(0) {}
};

ostream& operator<<(ostream& os, const UCT_node& n);

pair<position, UCT_node*> get_best_move(const UCT_node* node);

double do_play(const GoBoard& input_board);

class UCT {
private:
	GoBoard root_board;
public:
	vector<UCT_node*> nodes;
public:
	UCT(const GoBoard& board, double mod_komi = 0) {
		root_board = board;
		root_board.komi += mod_komi;
		nodes.push_back(new UCT_node(NULL));
	}

	~UCT() {
		for (unsigned int i = 0; i < nodes.size(); ++i) {
			delete nodes[i];
		}
	}

	void print_node_tree() const;

	double evaluateBoard() {
		UCT_node* node = nodes[0];
		GoBoard board(root_board);
		double score = play_simulation(board, node);
// 		cout << " score ==> " << score << endl;
// 		print_node_tree(node);
		return score;
	}

	const UCT_node* getRootNode() const {
		return nodes[0];
	}

	void addLegalMoveChildNodes(GoBoard& board, UCT_node* n) {
		vector<position> posses = board.getInterestingPlays();
		if (posses.size() == 0) {
			n->children.reserve(1);
			n->children.push_back(make_pair(position(-1, -1), (UCT_node*)NULL));
		} else {
			n->children.reserve(posses.size());
			for (unsigned int i = 0; i < posses.size(); ++i) {
				n->children.push_back(make_pair(posses[i], (UCT_node*)NULL));
			}
		}
	}

	int get_best_child_idx(UCT_node* n, const GoBoard& board) const;

	double estimate_score(const GoBoard& board, const UCT_node* n) const {
		return do_play(board);
	}

	double get_win_value(double score) const {
// 		if (score > 20) return 1.2;
// 		if (score < -20) return -0.2;
// 		if (score > 5) return 1.1;
// 		if (score < -5) return -0.1;
// 		if (score < 0) return 0.3;
// 		if (score > 0) return 0.7;
		if (score < 0) return 0.0;
		if (score > 0) return 1.0;
		return 0.5;
	}

	double play_simulation(GoBoard& board, UCT_node* n) {
		// Our move is already played on the board
		int player = 3 - board.player;
		if (board.is_game_over()) {
			double gamescore = board.get_score();
			n->plays++;
			if (gamescore > 0 && player == 1) n->wins++;
			else if (gamescore < 0 && player == 2) n->wins++;
			return gamescore;
		} else if (n->plays >= 1) {
			if (!n->children.size()) {
				addLegalMoveChildNodes(board, n);
			}
			int bestidx = get_best_child_idx(n, board);
			UCT_node* bestChild = n->children[bestidx].second;
			if (!bestChild) {
				bestChild = new UCT_node(n);
				n->children[bestidx].second = bestChild;
				nodes.push_back(bestChild);
			}
			board.play(n->children[bestidx].first);
			double gamescore = play_simulation(board, bestChild);
			n->plays++;
			if (gamescore > 0 && player == 1) n->wins++;
			else if (gamescore < 0 && player == 2) n->wins++;
			return gamescore;
		} else {
// 			double one = estimate_score(board, n);
			double two = estimate_score(board, n);
// 			double three = estimate_score(board, n);
// 			if (two < one) swap(one, two);
// 			if (three < two) swap(two, three);
// 			if (two < one) swap(one, two);
// 			assert(one <= two);
// 			assert(two <= three);
			n->plays++;
			double gamescore = two;
			if (player == 1) {
				n->wins += get_win_value(two);
			} else if (player == 2) {
				n->wins += get_win_value(-two);
			}
// 			board.print();
// 			cout << " ==> gamescore = " << gamescore << endl << endl;
			return gamescore;
		}
	}

	position get_best_move() const;
};

}

#endif
