#include "common.h"
#include "extra.h"
#include "sgf.h"
#include "influence.h"

#include <bitset>

extern vector<game_record> influence_map;

map<unsigned int, position> cbr_hash_position_map;

namespace go {

double get_position_difference_bias(const position& newpos, const position& oldpos) {
	if (newpos == oldpos) return 0.4;
	double diffx = newpos.x - oldpos.x;
	double diffy = newpos.y - oldpos.y;
	if (diffx * diffx + diffy * diffy <= 1) return 0.3;
	if (diffx * diffx + diffy * diffy <= 2) return 0.2;
	if (diffx * diffx + diffy * diffy <= 5 * 5) return 0.1;
	return 0.0;
}

position get_cbr_closest_position(const GoBoard& board) {
	map<unsigned int, position>::const_iterator it = cbr_hash_position_map.find(board.hash_code);
	if (it == cbr_hash_position_map.end()) {
		position o(-1, -1);

		bitset<3 * board_size * board_size> code = get_influence_code(board);

		int best_match = -100;
		for (unsigned int i = 0; i < influence_map.size(); ++i) {
			if (influence_map[i].player != board.player) continue;
			int goodness = (influence_map[i].code & code).count();
			if (goodness > best_match) {
				best_match = goodness;
				o = influence_map[i].pos;
			}
		}
		cbr_hash_position_map[board.hash_code] = o;
		return o;
	}
	return it->second;
}

string get_cbr_source_filename(const GoBoard& board) {
	string fn = "";

	bitset<3 * board_size * board_size> code = get_influence_code(board);

	int best_match = -100;
	for (unsigned int i = 0; i < influence_map.size(); ++i) {
		if (influence_map[i].player != board.player) continue;
		int goodness = (influence_map[i].code & code).count();
		if (goodness > best_match) {
			best_match = goodness;
			fn = influence_map[i].filename;
		}
	}
	for (int i = fn.size() - 1; i >= 0; --i) {
		if (fn[i] == '/') {
			string newfn = "";
			for (int j = i + 1; j < int(fn.size()); ++j) {
				newfn += fn[j];
			}
			fn = newfn;
		}
	}
	return fn;
}

double get_cbr_bias(const position& p, const GoBoard& board) {
	position o = get_cbr_closest_position(board);
	return get_position_difference_bias(p, o);
}

GoBoard get_funkiest_board(const vector<pair<bitset<3 * board_size * board_size>, GoBoard > >& influence_map, const bitset<3 * board_size * board_size>& code) {
	GoBoard board;
	int best_match = 0;
	for (unsigned int i = 0; i < influence_map.size(); ++i) {
		int goodness = (influence_map[i].first & code).count();
		if (goodness > best_match) {
			best_match = goodness;
			board = influence_map[i].second;
		}
	}
	return board;
}

void add_sgf_file(string filename, vector<game_record>& influence_map) {
	vector<GoBoard> boards = load_sgf_file(filename);
	for (int i = 0; i < int(boards.size()) - 1; ++i) {
		position play = boards[i + 1].get_last_play();
		int player = boards[i].player;
		assert(play.is_move());
		bitset<3 * board_size * board_size> code = get_influence_code(boards[i]);
// 		cout << "adding " << boards[i].hash_code << " -> " << play << " mapping" << player << endl;
		game_record gr = { code, play, player, filename };
		influence_map.push_back(gr);
// 		break;
	}
}

position get_player_input_move(const GoBoard& board) {
	while (true) {
		char x;
		int y;
		cout << "Enter move: " << flush;
		cin >> x >> y;
		cin.clear();
		cin.ignore(10000, '\n');
		if (x == 'x') {
			return position(-2, -2);
		}
		cout << "playing " << x << " " << y << endl;
		go::position p;
		if (x >= 'a' && x <= 'a' + board_size + 1) {
			p.x = x - 'a';
			if (x > 'i') p.x--;
		} else if (x >= 'A' && x <= 'A' + board_size + 1) {
			p.x = x - 'A';
			if (x > 'I') p.x--;
		} else {
			cout << "unknown move." << endl;
			continue;
		}
		if (y >= 1 && y <= board_size) {
			p.y = y - 1;
		} else {
			cout << "unknown move." << endl;
			continue;
		}
		if (board.getTile(p)) {
			cout << "illegal move." << endl;
			continue;
		}
		return p;
	}
}

ostream& operator<<(ostream& os, const UCT_node& n) {
// 	os << hex << n.board.hash_code << dec << " ";
	os << "wins[" << n.wins << "] / plays[" << n.plays << "] = " << (double(n.wins) / n.plays);
	return os;
}

}

void dump_winrate_file(const go::UCT_node* n) {
	double rates[board_size][board_size];
	memset(rates, 0, sizeof(rates));
	ofstream ofs("winrate-visualization.data");

	for (unsigned int i = 0; i < n->children.size(); ++i) {
		pair<go::position, go::UCT_node*> it = n->children[i];
		rates[it.first.y][it.first.x] = double(it.second->wins) / it.second->plays;
	}

	ofs << "[" << endl;
	for (int y = board_size - 1; y >= 0; --y) {
		ofs << "  [";
		for (int x = 0; x < board_size; ++x) {
			ofs << fixed << rates[y][x] << ",   ";
		}
		ofs << "]," << endl;
	}
	ofs << "]" << endl;
}
