#include "common.h"
#include "influence.h"
#include "GoBoard.h"

#include <bitset>

namespace go {

void spread_influence_area(const GoBoard& board, position p_start, board_influence& influence) {
	int t = board.getTile(p_start);
	if (t == 0) return;
	assert(t == 1 || t == 2);
	for (int j = -5; j < 5; ++j) {
		for (int i = -5; i < 5; ++i) {
			int x = p_start.x + i;
			int y = p_start.y + j;
			if (y < 0 || y >= board_size) continue;
			if (x < 0 || x >= board_size) continue;
			double dist = sqrt(i * i + j * j);
			double weight = 0;
			if (dist < 0.9) {
				weight = 30.0;
			} else if (dist < 5) {
				weight = 1.0 / dist;
			}
			influence.get(t, position(x, y)) += weight;
		}
	}
}

void spread_influence_flow(const GoBoard& board, position p_start, board_influence& influence) {
	int t = board.getTile(p_start);
	if (t == 0) return;
	assert(t == 1 || t == 2);
	deque<pair<position, int> > q;
	q.push_back(make_pair(p_start, 5));
	int offx[] = { -1, +1, 0, 0 };
	int offy[] = { 0, 0, -1, +1 };
	influence.get(t, p_start) += 2.0;
	while (q.size()) {
		position p = q.front().first;
		int counter = q.front().second;
		q.pop_front();
		influence.get(t, p) += counter / 5.0;
		counter--;
		if (counter == 0) continue;
		for (int i = 0; i < 4; ++i) {
			int x = p.x + offx[i];
			int y = p.y + offy[i];
			if (x < 0 || x >= board_size) continue;
			if (y < 0 || y >= board_size) continue;
			if (board.getTile(position(x, y))) continue;
			else q.push_back(make_pair(position(x, y), counter));
		}
	}
}

board_influence get_board_influence(const GoBoard& board) {
	board_influence influence;
	for (int y = 0; y < board_size; ++y) {
		for (int x = 0; x < board_size; ++x) {
			position p(x, y);
// 			spread_influence_area(board, p, influence);
			spread_influence_flow(board, p, influence);
// 			influence[1][2][3] = 70;
		}
	}
	return influence;
}

void print_board_influence(const GoBoard& board, double multiplier) {
	board_influence influence = get_board_influence(board);
	cout << "   ";
	for (int x = 0; x < board_size; ++x) {
		char c = char('A' + x);
		if (c >= 'I') c++;
		cout << c << " ";
	}
	cout << endl;
	for (int y = board_size - 1; y >= 0; --y) {
		if (y + 1 < 10) cout << " ";
		cout << (y + 1) << " ";
		for (int x = 0; x < board_size; ++x) {
			char c = ' ';
			position p(x, y);
			double sum_white = influence.get(2, p) - influence.get(1, p) * 2;
			double sum_black = influence.get(1, p) - influence.get(2, p) * 2;
			sum_white *= multiplier;
			sum_black *= multiplier;
			if (sum_white >= 2) c = 'O';
			else if (sum_black >= 2) c = 'X';
			else if (sum_white >= 0.2) c = 'o';
			else if (sum_black >= 0.2) c = 'x';
// 			cout << influence[1][y][x] << "," << influence[2][y][x] << "   ";
			cout << c << " ";
		}
		cout << " " << (y + 1) << endl;
	}
	cout << "   ";
	for (int x = 0; x < board_size; ++x) {
		char c = char('A' + x);
		if (c >= 'I') c++;
		cout << c << " ";
	}
	cout << endl;
	cout << endl;
}

bitset<3 * board_size * board_size> get_influence_code(const GoBoard& board) {
	board_influence influence = get_board_influence(board);
	bitset<3 * board_size * board_size> bits;
	for (int y = 0; y < board_size; ++y) {
		for (int x = 0; x < board_size; ++x) {
			int r = 0;
			double sum = influence.get(2, position(x, y)) - influence.get(1, position(x, y));
			if (sum <= -0.2) r = 2;
			if (sum >= 0.2) r = 1;
			if (r == 0) {
				bits.set((y * board_size + x) * 3 + 0);
			} else if (r == 1) {
				bits.set((y * board_size + x) * 3 + 1);
			} else if (r == 2) {
				bits.set((y * board_size + x) * 3 + 2);
			}
		}
	}
// 	cout << bits.size() << ": " << bits << endl;
	return bits;
}

}
